//
// Icosahedron
//
// The 3D model can be rotated by the mouse movement
// with the left mouse button pressed.
//
// The program is based on the class GLWindow, that presents a
// convenient C++ interface to OpenGL and X11 functions.
//
//                                      Written by V.Borisenko
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <assert.h>
#include "GLWindow.h"
#include "R3Graph.h"
#include "icosahed.h"

using namespace R3Graph;

static void vertex(const R3Point& p);

//--------------------------------------------------
// Definition of class "MyWindow"
//
class MyWindow: public GLWindow {  // Our main class derived from GLWindow
public:
    Icosahedron icosahedron;
    bool wireframe;

    GLdouble    m_Alpha;    // Angle of rotation around vert.axis in degrees
    GLdouble    m_Beta;     // Angle of rotation around hor.axis in degrees
    I2Point     m_MousePos; // Previous position of mouse pointer

public:
    MyWindow():                 // Constructor
        icosahedron(),
        wireframe(true),
        GLWindow(),
        m_Alpha(0.),
        m_Beta(0.),
        m_MousePos(-1, -1)
    {}

    void render();                      // Draw a scene graph

    virtual void onExpose(XEvent& event);
    virtual void onKeyPress(XEvent& event);
    virtual void onButtonPress(XEvent& event);
    virtual void onButtonRelease(XEvent& event);
    virtual void onMotionNotify(XEvent& event);
};

//
// Process the Expose event: draw in the window
//
void MyWindow::onExpose(XEvent& /* event */) {
    makeCurrent();

    // Rotate the scene
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();

    glEnable(GL_LIGHT0);        // Turn on the first light source

    // Rotate a model to have a cabinet projection
    glRotatef(30., 1., 0., 0.);
    glRotatef(-30., 0., 1., 0.);

    // Go to mathematical coord. system: Z-axis goes up
    glRotatef(-90., 1., 0., 0.);

    // Rotate a model
    glRotatef(m_Beta, 1., 0., 0.);
    glRotatef(m_Alpha, 0., 0., 1.);

    render();           // Draw a scene graph

    swapBuffers();
}

//
// Process the KeyPress event: 
// if "q" is pressed, then close the window
//
void MyWindow::onKeyPress(XEvent& event) {
    KeySym key;
    char keyName[256];
    int nameLen = XLookupString(&(event.xkey), keyName, 255, &key, 0);
    printf("KeyPress: keycode=0x%x, state=0x%x, KeySym=0x%x\n",
        event.xkey.keycode, event.xkey.state, (int) key);
    if (nameLen > 0) {
        keyName[nameLen] = 0;
        printf("\"%s\" button pressed.\n", keyName);
        if (keyName[0] == 'q') { // quit => close window
            destroyWindow();
        } else if (keyName[0] == 's') { // s => split faces
            icosahedron.split();
            redraw();
        } else if (keyName[0] == 'w') { // wirefram off/on
            wireframe = !wireframe;
            redraw();
        } else if (keyName[0] == 'i') { // quit => close window
            icosahedron.initialize();
            redraw();
        } else {
            // Set initial position.
            m_Alpha = 0.;
            m_Beta = 0.;
            redraw();
        }
    }
}

// Process mouse click
void MyWindow::onButtonPress(XEvent& event) {
    // int x = event.xbutton.x;
    // int y = event.xbutton.y;
    int mouseButton = event.xbutton.button;

    // printf("Mouse click: x=%d, y=%d, button=%d\n", x, y, mouseButton);

    if (mouseButton != Button1) {
        // Set initial position on right and middle mouse buttons
        m_Alpha = 0.;
        m_Beta = 0.;
        redraw();
    } else {
        m_MousePos.x = (-1); m_MousePos.y = (-1); // Mouse position undefined
    }
}

void MyWindow::onButtonRelease(XEvent& /* event */) {
    m_MousePos.x = (-1); m_MousePos.y = (-1); // Mouse position undefined
}

//
// Process the mouse movement
//
void MyWindow::onMotionNotify(XEvent& event) {
    int x = event.xbutton.x;
    int y = event.xbutton.y;
    if ((event.xbutton.state & Button1Mask) != 0) {
        if (m_MousePos.x >= 0) {
            int dx = x - m_MousePos.x;
            int dy = y - m_MousePos.y;
            if (m_Beta > 100. || m_Beta < (-100.))
                dx = -dx;

            m_Alpha += (GLdouble) dx * 0.1;
            if (fabs(m_Alpha) > 360.)
                m_Alpha -= ((int) m_Alpha / 360) * 360.;

            m_Beta += (GLdouble) dy * 0.1;
            if (fabs(m_Beta) > 360.)
                m_Beta -= ((int) m_Beta / 360) * 360.;

            redraw();
        }
        m_MousePos.x = x; m_MousePos.y = y;
    } else {
        m_MousePos.x = (-1); m_MousePos.y = (-1);   // Undefined
    }
}

static const int NUM_COLORS = 20;
static const char * const COLORS[20] = {
    "red", "green", "blue", "brown", "ivory",
    "MistyRose", "SlateGray", "navy", "magenta", "SeaGreen",
    "cyan", "SteelBlue", "turquoise", "aquamarine", "LawnGreen",
    "IndianRed", "SaddleBrown", "wheat", "salmon", "orange"
    //... "chartreuse", "yellow", "OliveDrab", "gold", "RosyBrown"
};

//
// Draw a 3D model
//
void MyWindow::render() {
    glClearColor(0.2, 0.3, 0.5, 1.); // Background color: dark blue
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

    GLfloat color[4];
    color[3] = 1.;
    if (wireframe) {
        color[0] = 0.5;
        color[1] = 0.5;
        color[2] = 0.5;
        glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
        glBegin(GL_LINES);
    } else {
        glBegin(GL_TRIANGLES);
    }

    for (int i = 0; i < icosahedron.numFaces; ++i) {
        XColor c;
        if (!wireframe) {
            R3Vector n = (
                icosahedron.vertices[icosahedron.faces[i][1]] - 
                icosahedron.vertices[icosahedron.faces[i][0]]
            ).vectorProduct(
                icosahedron.vertices[icosahedron.faces[i][2]] - 
                icosahedron.vertices[icosahedron.faces[i][0]]
            );
            n.normalize();
            glNormal3d(n.x, n.y, n.z);

            const char *colorName = COLORS[i % NUM_COLORS];

            XParseColor(
                m_Display,
                DefaultColormap(m_Display, m_Screen),
                colorName,
                &c
            );
            const GLfloat MAX_COLOR_COMP = 65535.;
            color[0] = (GLfloat) c.red / MAX_COLOR_COMP,
            color[1] = (GLfloat) c.green / MAX_COLOR_COMP,
            color[2] = (GLfloat) c.blue / MAX_COLOR_COMP,
            glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE, color);
        }

        for (int j = 0; j < 3; ++j) {
            if (wireframe) {
                R3Point& v0 = icosahedron.vertices[
                    icosahedron.faces[i][j]
                ];
                R3Point& v1 = icosahedron.vertices[
                    icosahedron.faces[i][(j+1)%3]
                ];
                R3Vector n = (
                    (v0-icosahedron.center()) +
                    (v1-icosahedron.center())
                ).normalize();
                glNormal3f(n.x, n.y, n.z);
                vertex(v0);
                vertex(v1);
            } else {
                vertex(
                    icosahedron.vertices[
                        icosahedron.faces[i][j]
                    ]
                );
            }
        }

        /*...
        int* t = icosahedron.triangles[i];
        for (int j = 0; j < 3; ++j) {
            int ie = t[j];
            int sgn = 1;
            if (ie < 0) {
                ie = (-ie);
                sgn = (-1);
            }
            --ie;
            assert(0 <= ie && ie < icosahedron.numEdges);
            if (sgn >= 0) {
                vertex(
                    icosahedron.vertices[icosahedron.edges[ie][0]]
                );
                vertex(
                    icosahedron.vertices[icosahedron.edges[ie][1]]
                );
            } else {
                vertex(
                    icosahedron.vertices[icosahedron.edges[ie][1]]
                );
                vertex(
                    icosahedron.vertices[icosahedron.edges[ie][0]]
                );
            }
        }
        ...*/
    }
    glEnd();
}

/////////////////////////////////////////////////////////////
// Main: initialize X, create an instance of MyWindow class,
//       and start the message loop
int main() {
    // Initialize X stuff
    if (!GWindow::initX()) {
        printf("Could not connect to X-server.\n");
        exit(1);
    }

    int width = GWindow::screenMaxX()/2;
    int height = GWindow::screenMaxY()/2;
    double aspect = (double) width / (double) height;

    MyWindow w;
    w.icosahedron.initialize();
    w.createWindow(
        I2Rectangle(                    // Window frame rectangle:
            I2Point(10, 10),            //     left-top corner,
            width, height               //     width, height
        ),
        R2Rectangle(                    // Window coordinate rectangle
            R2Point(-2. * aspect, -2.), //     left-top corner,
            4. * aspect, 4.             //     width, height
        ),
        "OpenGL Test"                   // Window title
    );
    w.setBackground("lightGray");
    w.makeCurrent();
    GLWindow::initializeOpenGL();

    GWindow::messageLoop();

    GWindow::closeX();
    return 0;
}

static void vertex(const R3Point& p) {
    glVertex3d(p.x, p.y, p.z);
}
