//
// Icosahedron and Triangulation of Sphere
//
#include <cmath>
#include <cstring>
#include <cassert>
#include "R3Graph.h"
#include "icosahed.h"

using namespace R3Graph;

Icosahedron::Icosahedron():
    numVertices(0),
    vertices(0),

    numFaces(0),
    faces(0),
    triangles(0),

    numEdges(0),
    edges(0)
{}

Icosahedron::~Icosahedron() {
    delete[] vertices;
    delete[] faces;
    delete[] triangles;
    delete[] edges;
}

static const R3Point icosahedron_vertices[12] = {
    R3Point(GOLD_SECT, 0, -1.),
    R3Point(GOLD_SECT, 0, 1.),
    R3Point(-GOLD_SECT, 0, 1.),
    R3Point(-GOLD_SECT, 0, -1.),
    R3Point(1., -GOLD_SECT, 0.),
    R3Point(1., GOLD_SECT, 0.),
    R3Point(-1., GOLD_SECT, 0.),
    R3Point(-1., -GOLD_SECT, 0.),
    R3Point(0., 1., -GOLD_SECT),
    R3Point(0., 1., GOLD_SECT),
    R3Point(0., -1., GOLD_SECT),
    R3Point(0., -1., -GOLD_SECT)
};

static const Icosahedron::Face icosahedron_faces[20] = {
    { 0, 5, 1 },        // 0
    { 0, 1, 4 },        // 1
    { 0, 4, 11},        // 2
    { 0, 11, 8 },       // 3
    { 0, 8, 5 },        // 4

    { 2, 3, 7 },        // 5
    { 2, 7, 10},        // 6
    { 2, 10, 9 },       // 7
    { 2, 9, 6 },        // 8
    { 2, 6, 3 },        // 9

    { 1, 9, 10},        // 10
    { 1, 10, 4 },       // 11
    { 4, 10, 7 },       // 12
    { 4, 7, 11},        // 13
    { 11, 7, 3 },       // 14
    { 11, 3, 8 },       // 15
    { 8, 3, 6 },        // 16
    { 8, 6, 5 },        // 17
    { 5, 6, 9 },        // 18
    { 5, 9, 1 }         // 19
};

static const Icosahedron::Edge icosahedron_edges[30] = {
    { 0, 1 },           // 0
    { 0, 4 },           // 1
    { 0, 5 },           // 2
    { 0, 8 },           // 3
    { 0, 11 },          // 4

    { 1, 4 },           // 5
    { 1, 5 },           // 6
    { 1, 9 },           // 7
    { 1, 10 },          // 8

    { 2, 3 },           // 9
    { 2, 6 },           // 10
    { 2, 7 },           // 11
    { 2, 9 },           // 12
    { 2, 10 },          // 13

    { 3, 6 },           // 14
    { 3, 7 },           // 15
    { 3, 8 },           // 16
    { 3, 11 },          // 17

    { 4, 7 },           // 18
    { 4, 10 },          // 19
    { 4, 11 },          // 20

    { 5, 6 },           // 21
    { 5, 8 },           // 22
    { 5, 9 },           // 23

    { 6, 8 },           // 24
    { 6, 9 },           // 25

    { 7, 10 },          // 26
    { 7, 11 },          // 27

    { 8, 11 },          // 28

    { 9, 10 }           // 29
};

static const Icosahedron::Triangle icosahedron_triangles[20] = {
    // 0, 5, 1
    { 2+1, -(6+1), -(0+1) },

    // 0, 1, 4
    { 0+1, 5+1, -(1+1) },

    // 0, 4, 11
    { 1+1, 20+1, -(4+1) },

    // 0, 11, 8
    { 4+1, -(28+1), -(3+1) },

    // 0, 8, 5
    { 3+1, -(22+1), -(2+1) },

    // 2, 3, 7
    { 9+1, 15+1, -(11+1) },

    // 2, 7, 10
    { 11+1, 26+1, -(13+1) },

    // 2, 10, 9
    { 13+1, -(29+1), -(12+1) },

    // 2, 9, 6
    { 12+1, -(25+1), -(10+1) },

    // 2, 6, 3
    { 10+1, -(14+1), -(9+1) },

    // 1, 9, 10
    { 7+1, 29+1, -(8+1) },

    // 1, 10, 4
    { 8+1, -(19+1), -(5+1) },

    // 4, 10, 7
    { 19+1, -(26+1), -(18+1) },

    // 4, 7, 11
    { 18+1, 27+1, -(20+1) },

    // 11, 7, 3
    { -(27+1), -(15+1), 17+1 },

    // 11, 3, 8
    { -(17+1), 16+1, 28+1 },

    // 8, 3, 6
    { -(16+1), 14+1, 24+1 },

    // 8, 6, 5
    { -(24+1), -(21+1), 22+1 },

    // 5, 6, 9
    { 21+1, 25+1, -(23+1) },

    // 5, 9, 1
    { 23+1, -(7+1), 6+1 }
};

void Icosahedron::initialize() {
    delete[] vertices;
    numVertices = 12;
    vertices = new R3Point[numVertices];

    delete[] faces;
    numFaces = 20;
    faces = new Face[numFaces];

    delete[] triangles;
    triangles = new Triangle[numFaces];

    delete edges;
    numEdges = 30;
    edges = new Edge[numEdges];

    /*...
    vertices[0] = R3Point(GOLD_SECT, 0, -1.);
    vertices[1] = R3Point(GOLD_SECT, 0, 1.);
    vertices[2] = R3Point(-GOLD_SECT, 0, 1.);
    vertices[3] = R3Point(-GOLD_SECT, 0, -1.);

    vertices[4] = R3Point(1., -GOLD_SECT, 0.);
    vertices[5] = R3Point(1., GOLD_SECT, 0.);
    vertices[6] = R3Point(-1., GOLD_SECT, 0.);
    vertices[7] = R3Point(-1., -GOLD_SECT, 0.);

    vertices[8] = R3Point(0., 1., -GOLD_SECT);
    vertices[9] = R3Point(0., 1., GOLD_SECT);
    vertices[10] = R3Point(0., -1., GOLD_SECT);
    vertices[11] = R3Point(0., -1., -GOLD_SECT);

    // Faces
    faces[0][0] = 0; faces[0][1] = 5; faces[0][2] = 1;      // 0 5 1
    faces[1][0] = 0; faces[1][1] = 1; faces[1][2] = 4;      // 0 1 4
    faces[2][0] = 0; faces[2][1] = 4; faces[2][2] = 11;     // 0 4 11
    faces[3][0] = 0; faces[3][1] = 11; faces[3][2] = 8;     // 0 11 8
    faces[4][0] = 0; faces[4][1] = 8; faces[4][2] = 5;      // 0 8 5

    faces[5][0] = 2; faces[5][1] = 3; faces[5][2] = 7;      // 2 3 7
    faces[6][0] = 2; faces[6][1] = 7; faces[6][2] = 10;     // 2 7 10
    faces[7][0] = 2; faces[7][1] = 10; faces[7][2] = 9;     // 2 10 9
    faces[8][0] = 2; faces[8][1] = 9; faces[8][2] = 6;      // 2 9 6
    faces[9][0] = 2; faces[9][1] = 6; faces[9][2] = 3;      // 2 6 3

    faces[10][0] = 1; faces[10][1] = 9; faces[10][2] = 10;  // 1 9 10
    faces[11][0] = 1; faces[11][1] = 10; faces[11][2] = 4;  // 1 10 4
    faces[12][0] = 4; faces[12][1] = 10; faces[12][2] = 7;  // 4 10 7
    faces[13][0] = 4; faces[13][1] = 7; faces[13][2] = 11;  // 4 7 11
    faces[14][0] = 11; faces[14][1] = 7; faces[14][2] = 3;  // 11 7 3
    faces[15][0] = 11; faces[15][1] = 3; faces[15][2] = 8;  // 11 3 8
    faces[16][0] = 8;  faces[16][1] = 3; faces[16][2] = 6;  // 8 3 6
    faces[17][0] = 8;  faces[17][1] = 6; faces[17][2] = 5;  // 8 6 5
    faces[18][0] = 5;  faces[18][1] = 6; faces[18][2] = 9;  // 5 6 9
    faces[19][0] = 5;  faces[19][1] = 9; faces[19][2] = 1;  // 5 9 1

    // Edges
    edges[0][0] = 0; edges[0][1] = 1;     // 0 1
    edges[1][0] = 0; edges[1][1] = 4;     // 0 4
    edges[2][0] = 0; edges[2][1] = 5;     // 0 5
    edges[3][0] = 0; edges[3][1] = 8;     // 0 8
    edges[4][0] = 0; edges[4][1] = 11;    // 0 11

    edges[5][0] = 1; edges[5][1] = 4;     // 1 4
    edges[6][0] = 1; edges[6][1] = 5;     // 1 5
    edges[7][0] = 1; edges[7][1] = 9;     // 1 9
    edges[8][0] = 1; edges[8][1] = 10;    // 1 10

    edges[9][0] = 2; edges[9][1] = 3;     // 2 3
    edges[10][0] = 2; edges[10][1] = 6;   // 2 6
    edges[11][0] = 2; edges[11][1] = 7;   // 2 7
    edges[12][0] = 2; edges[12][1] = 9;   // 2 9
    edges[13][0] = 2; edges[13][1] = 10;  // 2 10

    edges[14][0] = 3; edges[14][1] = 6;   // 3 6
    edges[15][0] = 3; edges[15][1] = 7;   // 3 7
    edges[16][0] = 3; edges[16][1] = 8;   // 3 8
    edges[17][0] = 3; edges[17][1] = 11;  // 3 11

    edges[18][0] = 4; edges[18][1] = 7;   // 4 7
    edges[19][0] = 4; edges[19][1] = 10;  // 4 10
    edges[20][0] = 4; edges[20][1] = 11;  // 4 11

    edges[21][0] = 5; edges[21][1] = 6;   // 5 6
    edges[22][0] = 5; edges[22][1] = 8;   // 5 8
    edges[23][0] = 5; edges[23][1] = 9;   // 5 9

    edges[24][0] = 6; edges[24][1] = 8;   // 6 8
    edges[25][0] = 6; edges[25][1] = 9;   // 6 9

    edges[26][0] = 7; edges[26][1] = 10;  // 7 10
    edges[27][0] = 7; edges[27][1] = 11;  // 7 11

    edges[28][0] = 8; edges[28][1] = 11;  // 8 11

    edges[29][0] = 9; edges[29][1] = 10;  // 9 10
    ...*/

    R3Point center(0., 0., 0.);

    int i;
    for (i = 0; i < numVertices; ++i) {
        //... vertices[i] = icosahedron_vertices[i];
        // Normalization
        R3Vector v = icosahedron_vertices[i] - center;
        v.normalize();
        vertices[i] = center + v;
    }
    for (i = 0; i < numFaces; ++i) {
        for (int j = 0; j < 3; ++j) {
            faces[i][j] = icosahedron_faces[i][j];
            triangles[i][j] = icosahedron_triangles[i][j];
        }
    }
    for (i = 0; i < numEdges; ++i) {
        for (int j = 0; j < 2; ++j) {
            edges[i][j] = icosahedron_edges[i][j];
        }
    }
}

void Icosahedron::split() {
    int numNewVertices = numVertices + numEdges;
    R3Point* newVertices = new R3Point[numNewVertices];
    memmove(newVertices, vertices, numVertices*sizeof(R3Point));

    int numNewFaces = 4 * numFaces;
    Face* newFaces = new Face[numNewFaces];

    Triangle* newTriangles = new Triangle[numNewFaces];

    int numNewEdges = 2 * numEdges + 3 * numFaces;
    Edge* newEdges = new Edge[numNewEdges];

    R3Point center(0., 0., 0.);

    int i;
    int newVertexIdx = numVertices;
    for (i = 0; i < numEdges; ++i) {
        int iv0 = edges[i][0];
        int iv1 = edges[i][1];
        R3Vector v0 = vertices[iv0] - center;
        R3Vector v1 = vertices[iv1] - center;
        R3Vector v = (v0 + v1).normalize();
        newVertices[newVertexIdx] = center + v;
        newEdges[2*i][0] = iv0;
        newEdges[2*i][1] = newVertexIdx;
        newEdges[2*i+1][0] = newVertexIdx;
        newEdges[2*i+1][1] = iv1;
        ++newVertexIdx;
    }

    int currentNewEdge = 2 * numEdges;

    for (i = 0; i < numFaces; ++i) {
        Triangle& triangle = triangles[i];
        int e0 = triangle[0];
        int s0 = 1;
        if (e0 < 0) {
            s0 = (-1); e0 = (-e0);
        }
        --e0;
        assert(0 <= e0 && e0 < numEdges);

        int e1 = triangle[1];
        int s1 = 1;
        if (e1 < 0) {
            s1 = (-1); e1 = (-e1);
        }
        --e1;
        assert(0 <= e1 && e1 < numEdges);

        int e2 = triangle[2];
        int s2 = 1;
        if (e2 < 0) {
            s2 = (-1); e2 = (-e2);
        }
        --e2;
        assert(0 <= e2 && e2 < numEdges);

        // New edges on sides of triangles
        int e00 = 2*e0;
        int e01 = 2*e0 + 1;
        int e10 = 2*e1;
        int e11 = 2*e1 + 1;
        int e20 = 2*e2;
        int e21 = 2*e2 + 1;

        // Indices of triangle vertices
        int iv0, iv1, iv2;
        // Indices of new vertices
        int iv01, iv12, iv20;

        if (s0 > 0) {
            iv0 = edges[e0][0];
            iv1 = edges[e0][1];
            assert(
               iv0 == newEdges[2*e0][0] &&
               iv1 == newEdges[2*e0+1][1]
            );
            iv01 = newEdges[2*e0][1];
        } else {
            iv0 = edges[e0][1];
            iv1 = edges[e0][0];
            assert(
                iv1 == newEdges[2*e0][0] &&
                iv0 == newEdges[2*e0+1][1]
            );
            iv01 = newEdges[2*e0][1];
        }
        assert(0 <= iv0 && iv0 < numVertices);
        assert(0 <= iv1 && iv1 < numVertices);
        assert(0 <= iv01 && iv01 < numNewVertices);

        if (s1 > 0) {
            assert(
                iv1 == edges[e1][0] &&
                iv1 == newEdges[2*e1][0]
            );
            iv2 = edges[e1][1];
            iv12 = newEdges[2*e1][1];
        } else {
            assert(
                iv1 == edges[e1][1] &&
                iv1 == newEdges[2*e1+1][1]
            );
            iv2 = edges[e1][0];
            iv12 = newEdges[2*e1][1];
        }
        assert(0 <= iv2 && iv2 < numVertices);
        assert(0 <= iv12 && iv12 < numNewVertices);

        if (s2 > 0) {
            assert(
                iv2 == edges[e2][0] &&
                iv2 == newEdges[2*e2][0]
            );
            iv20 = newEdges[2*e2][1];
        } else {
            assert(
                iv2 == edges[e2][1] &&
                iv2 == newEdges[2*e2+1][1]
            );
            iv20 = newEdges[2*e2][1];
        }
        assert(0 <= iv20 && iv20 < numNewVertices);

        // Add 3 new edges
        newEdges[currentNewEdge][0] = iv01;
        newEdges[currentNewEdge][1] = iv12;
        int edge01 = currentNewEdge;
        ++currentNewEdge;

        newEdges[currentNewEdge][0] = iv12;
        newEdges[currentNewEdge][1] = iv20;
        int edge12 = currentNewEdge;
        ++currentNewEdge;

        newEdges[currentNewEdge][0] = iv20;
        newEdges[currentNewEdge][1] = iv01;
        int edge20 = currentNewEdge;
        ++currentNewEdge;

        // Add 4 new triangles
        // triangle iv0, iv01, iv20
        int tr_idx = i*4;
        newFaces[tr_idx][0] = iv0;
        newFaces[tr_idx][1] = iv01;
        newFaces[tr_idx][2] = iv20;

        if (s0 > 0) {
            newTriangles[tr_idx][0] = (2*e0) + 1;
        } else {
            newTriangles[tr_idx][0] = (-((2*e0+1) + 1));
        }
        newTriangles[tr_idx][1] = (-(edge20 + 1));
        if (s2 > 0) {
            newTriangles[tr_idx][2] = (2*e2+1) + 1;
        } else {
            newTriangles[tr_idx][2] = (-((2*e2) + 1));
        }

        // triangle iv01, iv12, iv20
        ++tr_idx;
        newFaces[tr_idx][0] = iv01;
        newFaces[tr_idx][1] = iv12;
        newFaces[tr_idx][2] = iv20;

        newTriangles[tr_idx][0] = edge01 + 1;
        newTriangles[tr_idx][1] = edge12 + 1;
        newTriangles[tr_idx][2] = edge20 + 1;

        // triangle iv01, iv1, iv12
        ++tr_idx;
        newFaces[tr_idx][0] = iv01;
        newFaces[tr_idx][1] = iv1;
        newFaces[tr_idx][2] = iv12;

        if (s0 > 0) {
            newTriangles[tr_idx][0] = (2*e0+1) + 1;
        } else {
            newTriangles[tr_idx][0] = (-((2*e0) + 1));
        }
        if (s1 > 0) {
            newTriangles[tr_idx][1] = (2*e1) + 1;
        } else {
            newTriangles[tr_idx][1] = (-((2*e1+1) + 1));
        }
        newTriangles[tr_idx][2] = (-(edge01 + 1));

        // triangle iv12, iv2, iv20
        ++tr_idx;
        newFaces[tr_idx][0] = iv12;
        newFaces[tr_idx][1] = iv2;
        newFaces[tr_idx][2] = iv20;

        if (s1 > 0) {
            newTriangles[tr_idx][0] = (2*e1+1) + 1;
        } else {
            newTriangles[tr_idx][0] = (-((2*e1) + 1));
        }
        if (s2 > 0) {
            newTriangles[tr_idx][1] = (2*e2) + 1;
        } else {
            newTriangles[tr_idx][1] = (-((2*e2+1) + 1));
        }
        newTriangles[tr_idx][2] = (-(edge12 + 1));
    } // end for (i = 0; i < numFaces; ++i)

    assert(currentNewEdge == numNewEdges);

    delete[] vertices;
    vertices = newVertices;
    numVertices = numNewVertices;

    delete[] faces;
    faces = newFaces;
    numFaces = numNewFaces;

    delete[] triangles;
    triangles = newTriangles;

    delete[] edges;
    edges = newEdges;
    numEdges = numNewEdges;
}

R3Vector Icosahedron::direction(int vertexIdx) const
{
    assert(
        0 <= vertexIdx && vertexIdx < numVertices &&
        vertices != 0
    );
    if (!(
        0 <= vertexIdx && vertexIdx < numVertices &&
        vertices != 0
    ))
        return R3Vector(1., 0., 0.);    // Random value
    else
        return (vertices[vertexIdx] - center());
}
