//
// Stack Calculator (non-graphic version)
//

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include "RealStack.h"
#include "R3Graph.h"
#include "tetrahed.h"

using namespace R3Graph;

static void onAdd();
static void onSub();
static void onMul();
static void onDiv();
static void onMod();
static void onPush(const char* line);
static void onPop();
static void onDup();
static void onExch();
static void onInit();
static void display();
static void printHelp();
static void onShow();

// Math functions
static void onSin();
static void onCos();
static void onTan();
static void onAsin();
static void onAcos();
static void onAtan();
static void onAtan2();
static void onToDeg();
static void onFromDeg();

static void onSqrt();
static void onPow();
static void onExp();
static void onLog();
static void onLog10();

// Constants
static void onPi();
static void onE();

// R3 Graphics
static void onR3Dist();
static void onR3Volume();
static void onR3Area();
static void onR3AngleVectors();
static void onR3AnglePlanes();

// Tetrahedron
static void onTetrahedronOuterRadius();
static void onTetrahedronInnerRadius();
static void onTetrahedronAspect();

static RealStack *stack = 0;

int main() {
    printHelp();

    char line[256];
    stack = new RealStack();

    // gets(line);
    scanf("%s", line);
    while (*line != 0) {
        int linelen = strlen(line);
        try {
            if (strcasecmp(line, "+") == 0)
                onAdd();
            else if (strcasecmp(line, "-") == 0)
                onSub();
            else if (strcasecmp(line, "*") == 0)
                onMul();
            else if (strcasecmp(line, "/") == 0)
                onDiv();
            else if (strcasecmp(line, "%") == 0)
                onMod();
            else if (strcasecmp(line, "=") == 0)
                display();
            else if (
                (linelen > 0 && isdigit(line[0])) ||
                (linelen > 1 && (line[0] == '-' || line[0] == '+') &&
                    isdigit(line[1]))
            )
                onPush(line);
            else if (strcasecmp(line, "pop") == 0)
                onPop();
            else if (strcasecmp(line, "dup") == 0)
                onDup();
            else if (strcasecmp(line, "exch") == 0)
                onExch();
            else if (strcasecmp(line, "init") == 0)
                onInit();
            else if (strcasecmp(line, "show") == 0)
                onShow();
            else if (strcasecmp(line, "sin") == 0)
                onSin();
            else if (strcasecmp(line, "cos") == 0)
                onCos();
            else if (strcasecmp(line, "tan") == 0)
                onTan();
            else if (strcasecmp(line, "asin") == 0)
                onAsin();
            else if (strcasecmp(line, "acos") == 0)
                onAcos();
            else if (strcasecmp(line, "atan") == 0)
                onAtan();
            else if (strcasecmp(line, "atan2") == 0)
                onAtan2();
            else if (strcasecmp(line, "sqrt") == 0)
                onSqrt();
            else if (strcasecmp(line, "pow") == 0)
                onPow();
            else if (strcasecmp(line, "log") == 0)
                onLog();
            else if (strcasecmp(line, "log10") == 0)
                onLog10();
            else if (strcasecmp(line, "exp") == 0)
                onExp();
            else if (strcasecmp(line, "pi") == 0)
                onPi();
            else if (strcasecmp(line, "e") == 0)
                onE();
            else if (strncasecmp(line, "r3dist", 6) == 0)
                onR3Dist();
            else if (strncasecmp(line, "r3vol", 5) == 0)
                onR3Volume();
            else if (strncasecmp(line, "r3are", 5) == 0)
                onR3Area();
            else if (strncasecmp(line, "r3anglev", 8) == 0)
                onR3AngleVectors();
            else if (strncasecmp(line, "r3anglep", 8) == 0)
                onR3AnglePlanes();
            else if (strncasecmp(line, "toDeg", 5) == 0)
                onToDeg();
            else if (strncasecmp(line, "fromDeg", 7) == 0)
                onFromDeg();
            else if (strncasecmp(line, "tetraOuterR", 6) == 0)
                onTetrahedronOuterRadius();
            else if (strncasecmp(line, "tetraInnerR", 6) == 0)
                onTetrahedronInnerRadius();
            else if (strncasecmp(line, "tetraAspect", 6) == 0)
                onTetrahedronAspect();
            else if (
                strcasecmp(line, "") == 0 ||
                (linelen > 0 && (line[0] == 'q' || line[0] == 'Q'))
            )
                break;
            else
                printHelp();
        } catch (StackException& e) {
            printf("Stack Exception: %s\n", e.reason);
        }
        // gets(line);
        scanf("%s", line);
    }

    delete stack;
    return 0;
}

static void onAdd() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(x + y);
    display();
}

static void onSub() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(x - y);
    display();
}

static void onMul() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(x * y);
    display();
}

static void onDiv() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(x / y);
    display();
}

static void onMod() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(fmod(x, y));
    display();
}

static void onPush(const char* line) {
    double x = atof(line);
    stack->push(x);
}

static void onPop() {
    stack->pop();
}

static void onInit() {
    stack->init();
}

static void display() {
    if (!stack->empty()) {
        printf("=%lf\n", stack->top());
    } else {
        printf("stack empty\n");
    }
}

static void onShow() {
    int d = stack->depth();
    printf("Depth of stack = %d.", d);
    if (d > 0)
        printf(" Stack elements:\n");
    else
        printf("\n");

    for (int i = 0; i < d; i++) {
        printf("\t%lf\n", stack->elementAt(i));
    }
}

static void onDup() {
    double x = stack->top();
    stack->push(x);
    display();
}

static void onExch() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(y);
    stack->push(x);
    display();
}

static void onSin() {
    double x = stack->pop();
    stack->push(sin(x));
    display();
}

static void onCos() {
    double x = stack->pop();
    stack->push(cos(x));
    display();
}

static void onTan() {
    double x = stack->pop();
    stack->push(tan(x));
    display();
}

static void onAsin() {
    double x = stack->pop();
    stack->push(asin(x));
    display();
}

static void onAcos() {
    double x = stack->pop();
    stack->push(acos(x));
    display();
}

static void onAtan() {
    double x = stack->pop();
    stack->push(atan(x));
    display();
}

static void onAtan2() {
    // Usage: atan2(y, x)
    double x = stack->pop();
    double y = stack->pop();
    stack->push(atan2(y, x));
    display();
}

static void onSqrt() {
    double x = stack->pop();
    stack->push(sqrt(x));
    display();
}

static void onPow() {
    double y = stack->pop();
    double x = stack->pop();
    stack->push(pow(x, y));
    display();
}

static void onExp() {
    double x = stack->pop();
    stack->push(exp(x));
    display();
}

static void onLog() {
    double x = stack->pop();
    stack->push(log(x));
    display();
}

static void onLog10() {
    double x = stack->pop();
    stack->push(log10(x));
    display();
}

static void onPi() {
    stack->push(M_PI);
    display();
}

static void onE() {
    stack->push(M_E);
    display();
}

static void onR3Dist() {
    double z2 = stack->pop(), y2 = stack->pop(), x2 = stack->pop();
    double z1 = stack->pop(), y1 = stack->pop(), x1 = stack->pop();
    stack->push(
        R3Point(x1, y1, z1).distance(R3Point(x2, y2, z2))
    );
    display();
}

static void onR3Volume() {
    double v3z = stack->pop(), v3y = stack->pop(), v3x = stack->pop();
    double v2z = stack->pop(), v2y = stack->pop(), v2x = stack->pop();
    double v1z = stack->pop(), v1y = stack->pop(), v1x = stack->pop();
    stack->push(
        R3Vector::volume(
            R3Vector(v1x, v1y, v1z),
            R3Vector(v2x, v2y, v2z),
            R3Vector(v3x, v3y, v3z)
        )
    );
    display();
}

static void onR3Area() {
    double v2z = stack->pop(), v2y = stack->pop(), v2x = stack->pop();
    double v1z = stack->pop(), v1y = stack->pop(), v1x = stack->pop();
    stack->push(
        R3Vector::area(
            R3Vector(v1x, v1y, v1z),
            R3Vector(v2x, v2y, v2z)
        )
    );
    display();
}

static void onTetrahedronOuterRadius() {
    double p3z = stack->pop(), p3y = stack->pop(), p3x = stack->pop();
    double p2z = stack->pop(), p2y = stack->pop(), p2x = stack->pop();
    double p1z = stack->pop(), p1y = stack->pop(), p1x = stack->pop();
    double p0z = stack->pop(), p0y = stack->pop(), p0x = stack->pop();
    stack->push(
        Tetrahedron(
            R3Point(p0x, p0y, p0z),
            R3Point(p1x, p1y, p1z),
            R3Point(p2x, p2y, p2z),
            R3Point(p3x, p3y, p3z)
        ).outerRadius()
    );
    display();
}

static void onTetrahedronInnerRadius() {
    double p3z = stack->pop(), p3y = stack->pop(), p3x = stack->pop();
    double p2z = stack->pop(), p2y = stack->pop(), p2x = stack->pop();
    double p1z = stack->pop(), p1y = stack->pop(), p1x = stack->pop();
    double p0z = stack->pop(), p0y = stack->pop(), p0x = stack->pop();
    stack->push(
        Tetrahedron(
            R3Point(p0x, p0y, p0z),
            R3Point(p1x, p1y, p1z),
            R3Point(p2x, p2y, p2z),
            R3Point(p3x, p3y, p3z)
        ).innerRadius()
    );
    display();
}

static void onTetrahedronAspect() {
    double p3z = stack->pop(), p3y = stack->pop(), p3x = stack->pop();
    double p2z = stack->pop(), p2y = stack->pop(), p2x = stack->pop();
    double p1z = stack->pop(), p1y = stack->pop(), p1x = stack->pop();
    double p0z = stack->pop(), p0y = stack->pop(), p0x = stack->pop();
    stack->push(
        Tetrahedron(
            R3Point(p0x, p0y, p0z),
            R3Point(p1x, p1y, p1z),
            R3Point(p2x, p2y, p2z),
            R3Point(p3x, p3y, p3z)
        ).aspect()
    );
    display();
}

static void onR3AngleVectors() {
    double z2 = stack->pop(), y2 = stack->pop(), x2 = stack->pop();
    double z1 = stack->pop(), y1 = stack->pop(), x1 = stack->pop();
    stack->push(
        R3Vector(x1, y1, z1).angle(
            R3Vector(x2, y2, z2)
        )
    );
    display();
}

static void onR3AnglePlanes() {
    // Axis (p0, p1); planes (p0, p1, p2), (p0, p1, p3).
    double z = stack->pop(), y = stack->pop(), x = stack->pop();
    R3Point p3(x, y, z);
    z = stack->pop(); y = stack->pop(); x = stack->pop();
    R3Point p2(x, y, z);
    z = stack->pop(); y = stack->pop(); x = stack->pop();
    R3Point p1(x, y, z);
    z = stack->pop(); y = stack->pop(); x = stack->pop();
    R3Point p0(x, y, z);
    R3Vector v = p1 - p0;
    R3Vector n1 = v.vectorProduct(p2 - p0);
    R3Vector n2 = v.vectorProduct(p3 - p0);
    stack->push(
        n1.angle(n2)
    );
    display();
}

static void onToDeg() {
    double x = stack->pop();
    stack->push(
        (x / M_PI) * 180.
    );
    display();
}

static void onFromDeg() {
    double x = stack->pop();
    stack->push(
        (x / 180.) * M_PI
    );
    display();
}

static void printHelp() {
    printf(
        "Stack Calculator commands:\n"
        "\t<number>\tPush to stack\n"
        "\t+, -, *, /, %\tAriphmetic operations\n"
        "\t\t\tMath. functions:\n"
        "\tsin, cos, tan,\n"
        "\tasin, acos, atan, atan2,\n"
        "\tsqrt, exp, log, log10, pow\n"
        "\tpi, e\t\tMath. constants\n"
        "\n"
        "\t=\t\tDisplay the stack top\n"
        "\tpop\t\tRemove the stack top\n"
        "\tdup\t\tDuplicate the stack top\n"
        "\texch\t\tExchange 2 elements at stack top\n"
        "\tshow\t\tShow the stack\n"
        "\tinit\t\tErase the stack\n"
        "\tquit\t\tEnd the program\n"
        "3D Graphics:\n"
        "\tr3dist\t\tdistance between points\n"
        "\tr3vol\t\tvolume of a box spanned on 3 vectors\n"
        "\tr3area\t\tarea of 2 vectors span\n"
        "\tr3anglevect\tangle between 2 vectors (radians)\n"
        "\tr3angleplanes\tangle between 2 planes\n"
        "\t\t\t\t(p1, p2, p3) and (p1, p2, p4) (radians)\n"
        "\ttoDeg\t\tconvert angle to degrees\n"
        "\tfromDeg\t\tconvert angle from degrees\n"
        "\ttetraOuterR\tradius of external sphere of tetrahedron\n"
        "\ttetraInnerR\tradius of internal sphere of tetrahedron\n"
        "\ttetraAspect\taspect of tetrahedron (outer radius / inner radius\n"
    );
}
